﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Anwendung                                                                           //
//  _________________________________                                                   //
//                                                                                      //
//  Copyright (c) 2019 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include <avr/pgmspace.h>
#include "Cy4Net.h"
#include "Cy4TypePlate.h"
#include "Storing.h"
#include "Board.h"
#include "Timer.h"
#include "Application.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static bool infoLedOn = false;

//////////////////////////////////////////////////////////////////////////////////////////
//  Cy4-Nachrichtenausstatter  ///////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
SetupInfoLedStateMessage(Cy4Message* message)
{
    Cy4Address address;
    UInt32     location;
    STORING_Load(STO_AutoPushMessageAddress , &address);
    STORING_Load(STO_AutoPushMessageLocation, &location);
    UInt08 data[1] = { infoLedOn };
    CY4MESSAGE_SetupWrite(message, address, location, data, sizeof(data));
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Cy4-Dienst Antwortbehandler  /////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnResponseOfInfoLedState(const Cy4Message* message)
{
    if (CY4MESSAGE_IsValid(message))
    {   // Übertragung erfolgreich. Beim nächsten Aufruf, LED-Zustand invertieren
        infoLedOn = !infoLedOn;
        // ...
    }
    else
    {   // Übertragung fehlgeschlagen. Anfrage gegebenenfalls wiederholen
        //
        // attempt++;
        // if (attempt < MAX_ATTEMPTS)
        // {   CY4NET_Post(SetupInfoLedStateMessage, OnResponseOfInfoLedState);
        // }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Cy4-Dienst Anfragebearbeiter  ////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: InfoLedState (read write)                                                   //
//                                                                                      //
//  Beschreibung: Liefert oder Ändert den Zustand der Info-LED.                         //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |INF|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [INF] Zustand der Info-LED: 0 = Aus, 1 = Ein.                                       //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnInfoLedState(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    if (CY4MESSAGE_Access(message) == CY4MESSAGE_READ)
    {   CY4MESSAGE_Data(message)[0] = BOARD_InfoLed();
    }
    else
    {   UInt08 data = CY4MESSAGE_ConstData(message)[0];
		if (data<2) BOARD_SetInfoLed((bool)data);		
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: PeriodicPushMessage (read write)                                            //
//                                                                                      //
//  Beschreibung: Zieladresse und Wiederholintervall der periodisch zu sendenden        //
//                Push-Nachricht.                                                       //
//                                                                                      //
//              0   1   2   3   4   5                                                   //
//            +---+---+---+---+---+---+                                                 //
//  Belegung: |INT|NET|DEV|LC0|LC1|LC2|                                                 //
//            +---+---+---+---+---+---+                                                 //
//                                                                                      //
//  [INT] Intervall für das Versenden der Push-Nachrichten: 1 .. 100 [s÷10], 0 = Aus    //
//  [NET] Push-Nachricht Ziel, Netzadresse  : 0 .. 255                                  //
//  [DEV] Push-Nachricht Ziel, Geräteadresse: 0 .. 255                                  //
//  [LCx] Push-Nachricht Ziel, Dienstort    : 0 .. FFFFFFh                              //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnPeriodicPushMessage(const Cy4Service* service, Cy4Message* message)
{
    // Parameter laden
    UInt08     interval;
    Cy4Address address;
    UInt32     location;
    STORING_Load(STO_AutoPushMessageInterval, &interval);
    STORING_Load(STO_AutoPushMessageAddress , &address);
    STORING_Load(STO_AutoPushMessageLocation, &location);
    // Dienst-Datenbereich bestücken
    UInt08  data[6];
    data[0] = interval;
    data[1] = CY4ADDRESS_Net(&address);
    data[2] = CY4ADDRESS_Dev(&address);
    data[3] = (UInt08)(location >> 16);
    data[4] = (UInt08)(location >>  8);
    data[5] = (UInt08)(location >>  0);
    // Daten gemäßt Zugriffsart austauschen
    CY4SERVICE_TransferPayload(service, message, data);
    // ... bei Leseanfragen könnte die Funktion hier beendet werden
    // Parameter wieder setzen
    interval = data[0] <= 100 ? data[0] : 100;  // Maximal 100
    address  = CY4ADDRESS_Create(data[1], data[2]);
    location = ((UInt32)data[3] << 16) + ((UInt16)data[4] << 8) + data[5];
    // Parameter speichern
    STORING_Save(&interval, STO_AutoPushMessageInterval);
    STORING_Save(&address , STO_AutoPushMessageAddress);
    STORING_Save(&location, STO_AutoPushMessageLocation);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: SystemLedTask (read write)                                                  //
//                                                                                      //
//  Beschreibung: Liefert oder ändert die Aufgabe der System-LED. Die Einstellung wird  //
//  gespeichert.                                                                        //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |TSK|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [TSK] Aufgabe der System-LED: 0 = Keine (LED aus)                                   //
//                                1 = LED ein                                           //
//                                2 = LED zeigt Busaktivität                            //
//                                3 = LED zeigt Herzschlagsignal                        //
//                                4 = LED zeigt Busaktivität und Herzschlagsignal       //
//                                x = Keine (LED aus)                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnSystemLedTask(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    if (CY4MESSAGE_Access(message) == CY4MESSAGE_READ)
    {   SysLedTask task;
        STORING_Load(STO_SysLedTask, &task);
        CY4MESSAGE_Data(message)[0] = (UInt08)task;
    }
    else
    {   SysLedTask task = (SysLedTask)CY4MESSAGE_ConstData(message)[0];
        if ((UInt08)task > 4) task = SYSLEDTASK_NONE;
        STORING_Save(&task, STO_SysLedTask);
        BOARD_SetSysLedTask(task);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: DeviceLabel (read write)                                                    //
//                                                                                      //
//  Beschreibung: Benutzerdefinierte Gerätebezeichnung.                                 //
//                                                                                      //
//              0   1   2   ...  59  60                                                 //
//            +---+---+---+-   -+---+---+                                               //
//  Belegung: |N01|N02|N03| ... |N60| 0 |                                               //
//            +---+---+---+-   -+---+---+                                               //
//                                                                                      //
//  [Nxx] Zeichen der allgemeinen Gerätebezeichnung: 0 .. 255 (ASCII/UTF-8)             //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnDeviceLabel(const Cy4Service* service, Cy4Message* message)
{
    UInt08  label[61];

    STORING_Load(STO_DeviceLabel, label);
    CY4SERVICE_TransferPayload(service, message, label);
    label[sizeof(label)-1] = '\0';
    STORING_Save(label, STO_DeviceLabel);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Dienst: FactorySettings (write only)                                                //
//                                                                                      //
//  Beschreibung: Zurückstellen aller Einstellungen auf den Ursprungszustand. Das       //
//  Board wird danach neu gestartet.                                                    //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |FST|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [FST] Auslöser zur Wiederherstellung der Werksvoreinstellungen: 5Eh ('^').          //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnFactorySettings(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E)
    {   STORING_RestoreDefaults();
        BOARD_Reset();
    }
    else
    {   CY4MESSAGE_Invalidate(message);
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: CallTypePlate (read only)                                            //
//                                                                                      //
//  Beschreibung: Geräte-Typenschild erfragen.                                          //
//                                                                                      //
//              0   1   2   3   4   5   6                                               //
//            +---+---+---+---+---+---+---+                                             //
//  Belegung: |AT2|AT1|AT0|AVE|BT2|BT1|BT0|                                             //
//            +---+---+---+---+---+---+---+                                             //
//                                                                                      //
//  [ATx] Anwendungstypennummer: Axxxxxh, xxxxx in BCD: 00001 .. 99999                  //
//  [AVE] Anwendungsversion:     01 .. 99, Haupt/Unternummer in BCD (Bsp: 12h = V1.2)   //
//  [BTx] Boardtypennummer:      Bxxxxxh, xxxxx in BCD: 00001 .. 99999                  //
//                                                                                      //
//                                                                                      //
//      Beispiel:                                                                       //
//                                                                                      //
//          +---+---+---+---+---+---+---+                                               //
//          |A0h|01h|23h|45h|B0h|67h|89h|      Anwendungstypennummer: 123               //
//          +---+---+---+---+---+---+---+      Anwendungsversion:     4.5               //
//                                             Boardtypennummer:      6789              //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnCallTypePlate(const Cy4Service* service, Cy4Message* message)
{
    static const UInt08  typePlate[][7] PROGMEM = {  { 'C','y','4','C','o','d','e' },
                                                     CY4TYPEPLATE(DEV_APPL_TYPE,
                                                                  DEV_APPL_VERS,
                                                                  DEV_BOARD_TYPE)
                                                  };

    UInt08 j = (UInt08)(CY4MESSAGE_Location(message) - CY4SERVICE_Base(service));
    for (UInt08 i=0; i<CY4MESSAGE_Count(message); i++)
    {   CY4MESSAGE_Data(message)[i] = pgm_read_byte(&typePlate[1][j]);
        j++;
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: EnterApplication (write only)                                        //
//                                                                                      //
//  Beschreibung: Startet die Anwendung neu.                                            //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |EAP|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [EAP] Auslöser zum (Neu-)Start der Applikation (Reset): 5Eh ('^') .                 //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnEnterApplication(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E) BOARD_Reset();
    else                                          CY4MESSAGE_Invalidate(message);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: EnterProgramLoader (write only)                                      //
//                                                                                      //
//  Beschreibung: Startet das System neu und kehrt im Programmlader-Modul zurück.       //
//                                                                                      //
//              0                                                                       //
//            +---+                                                                     //
//  Belegung: |EPL|                                                                     //
//            +---+                                                                     //
//                                                                                      //
//  [EPL] Auslöser zum Eintritt in den Programmlader: 5Eh ('^') .                       //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnEnterProgramLoader(const Cy4Service* service, Cy4Message* message)
{
    (void)service;
    if (CY4MESSAGE_ConstData(message)[0] == 0x5E) BOARD_EnterPgmLdr();
    else                                          CY4MESSAGE_Invalidate(message);
}

//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Pflichtdienst: ChangeCy4Address (write only)                                        //
//                                                                                      //
//  Beschreibung: Wechsel der Geräteadresse.                                            //
//                                                                                      //
//              0   1                                                                   //
//            +---+---+                                                                 //
//  Belegung: |NET|DEV|                                                                 //
//            +---+---+                                                                 //
//                                                                                      //
//  [NET] Neue Netzadresse:   1 .. 255                                                  //
//  [DEV] Neue Geräteadresse: 1 .. 255                                                  //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
static void
OnChangeCy4Address(const Cy4Service* service, Cy4Message* message)
{
    UInt08  data[2];
    {   Cy4Address address = CY4NET_Address();
        data[0] = CY4ADDRESS_Net(&address);
        data[1] = CY4ADDRESS_Dev(&address);
    }
    CY4SERVICE_TransferPayload(service, message, data);
    {   Cy4Address address = CY4ADDRESS(data[0], data[1]);
        if (CY4ADDRESS_IsBroadcast(&address) || !CY4ADDRESS_IsValid(&address))
        {   CY4MESSAGE_Invalidate(message);
        }
        else
        {   CY4NET_SetAddress(address);
            STORING_SaveCy4SelfAdr(&address);
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Prozessaufrufe  //////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Startup(void)
{
    static const Cy4ServiceItem  servicesTable[] PROGMEM =
    {   //|                    service                     |        handler        |
        //+------------------+-----+-----------------------+-----------------------+
        { CY4SERVICE(0x100200,  1  , CY4SERVICE_READ_WRITE), OnInfoLedState        },
        { CY4SERVICE(0x100400,  6  , CY4SERVICE_READ_WRITE), OnPeriodicPushMessage },
        { CY4SERVICE(0xFFF020,  1  , CY4SERVICE_READ_WRITE), OnSystemLedTask       },
        { CY4SERVICE(0xFFFF00, 61  , CY4SERVICE_READ_WRITE), OnDeviceLabel         },
        { CY4SERVICE(0xFFFF40,  1  , CY4SERVICE_WRITE_ONLY), OnFactorySettings     },
        { CY4SERVICE(0xFFFFF0,  7  , CY4SERVICE_READ_ONLY ), OnCallTypePlate       },
        { CY4SERVICE(0xFFFFF8,  1  , CY4SERVICE_WRITE_ONLY), OnEnterApplication    },
        { CY4SERVICE(0xFFFFFA,  1  , CY4SERVICE_WRITE_ONLY), OnEnterProgramLoader  },
        { CY4SERVICE(0xFFFFFC,  2  , CY4SERVICE_WRITE_ONLY), OnChangeCy4Address    }
    };
    CY4NET_AssignServices(servicesTable);
    Cy4Address address = STORING_LoadCy4SelfAdr();
    CY4NET_SetAddress(address);
    SysLedTask task;
    STORING_Load(STO_SysLedTask, &task);
    BOARD_SetSysLedTask(task);
}

//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Proceed(void)
{
    UInt08 interval;
    STORING_Load(STO_AutoPushMessageInterval, &interval);
    if (interval != 0)
    {   // Periodisch Push-Nachrichten senden
        UInt16 millisecs = interval*100;  // Einheit von <interval> ist [s÷10]
        static Timer timer = TIMER_MILLISECONDS(0);
        if (millisecs != TIMER_Duration(&timer)) TIMER_Start(&timer, millisecs);
        if (TIMER_RestartIfExpired(&timer))
        {   if (!CY4NET_IsPending(SetupInfoLedStateMessage, OnResponseOfInfoLedState))
            {   CY4NET_Post(SetupInfoLedStateMessage, OnResponseOfInfoLedState);
            }
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
void
APPLICATION_Shutdown(void)
{
}

//////////////////////////////////////////////////////////////////////////////////////////
