﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Passiver Zeitgeber                                                                  //
//  _________________________________                                                   //
//                                                                                      //
//  Copyright (c) 2007 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Beschreibung:                                                                       //
//                                                                                      //
//  Timer stellt einen monotonen "Single-Shot"-Zeitgeber zur Verfügung. Der Stand des   //
//  Zeitgebers wird nicht selbstinitiiert durch ein Ereignis oder einen Rückruf mitge-  //
//  teilt, sondern muss durch sukzessive Aufrufe der Timer-Zugriffsfunktionen erfragt   //
//  werden (passiv). Um dabei Überläufe zu vermeiden, dürfen Aufrufe bei laufendem      //
//  Zeitgeber nicht länger als eine Minute auseinanderliegen.                           //
//  Das Zeitgeber-Modul benötigt folgende Funktionen, die extern zur Verfügung gestellt //
//  werden müssen:                                                                      //
//                            UInt16   BOARD_MSecTicks(void)                            //
//                            UInt16   BOARD_SecTicks(void)                             //
//                                                                                      //
//  Die zeitliche Auflösung dieser Funktionen bestimmen auch die Genauigkeit der Zeit-  //
//  geber. Ein auf n Zeiteinheiten gestellter Zeitgeber läuft in der Zeit zwischen      //
//  (n-1) und n Zeiteinheiten ab. Zur Handhabung des Zeitgebers s.a. Timer_Start().     //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////| RG |//

#include "Timer.h"
#include "Board.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Lokale Funktionen  ///////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static inline UInt16
Ticks(Timer* timer)
{
    switch (timer->base)
    {   case TIMERBASE_MILLISECONDS: return BOARD_MSecTicks();
        case TIMERBASE_SECONDS:      return BOARD_SecTicks();
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Funktionen  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////////////////
//  TIMER_Start(): Startet den Zeitgeber mit der angegebenen Dauer. Ein gegebenenfalls  //
//  -------------  noch laufender Zeitgeber wird mit Aufruf dieser Funktion zurück-     //
//  gesetzt und mit der angegebenen Dauer neu gestartet. Die Festlegung der Zeitbasis   //
//  wird bei Instanziierung der TIMER-Struktur oder durch Aufruf von TIMER_SetBase()    //
//  getroffen. Zur Auswahl stehen Sekunden oder Millisekunden.                          //
//                                                                                      //
//   Beispiel:   Timer  secTimer  = TIMER_SECONDS(123);                                 //
//               Timer  msecTimer = TIMER_MILLISECONDS(500);                            //
//                                                                                      //
//  Hinweis: Der initiale Zustand nach Instanziierung eines Zeitgebers ist stets        //
//           "abgelaufen". Erst ein Aufruf von TIMER_Start(), TIMER_Restart() oder      //
//           TIMER_RestartIfExpired() setzt ihn in Gang.                                //
//                                                                                      //
//  Eingabe: · timer:    Zeiger auf die Zeitgeber-Struktur.                             //
//           · duration: Dauer des Zeitgebers: 0 .. 15000 [s|ms].                       //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
TIMER_Start(Timer* timer, UInt16 duration)
{
    if (duration > 15000) duration = 15000;
    timer->duration = duration;
    TIMER_Restart(timer);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  TIMER_Restart(): Startet den Zeitgeber erneut. Die Dauer entspricht der im letzten  //
//  ---------------  TIMER_Start()-Aufruf angegebenen oder der per TIMER_..()-Initia-   //
//  lisierung festgelegten Dauer. Ein eventuell noch laufender Zeitgeber wird mit Auf-  //
//  ruf dieser Funktion wieder zurückgesetzt und neu gestartet.                         //
//                                                                                      //
//  Eingabe: · timer: Zeiger auf die Zeitgeber-Struktur.                                //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
TIMER_Restart(Timer* timer)
{
    timer->hasExpired = timer->duration == 0;
    timer->ticksStamp = Ticks(timer);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  TIMER_Stop(): Stoppt den Zeigeber. Der Zeitgeber ist unmittelbar nach Aufruf der    //
//  ------------  Funktion abgelaufen (s.a. TIMER_HasExpired()).                        //
//                                                                                      //
//  Eingabe: · timer: Zeiger auf die Zeitgeber-Struktur.                                //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
TIMER_Stop(Timer* timer)
{
    timer->hasExpired = true;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  TIMER_HasExpired(): Prüft, ob der Zeitgeber abgelaufen ist.                         //
//  ------------------                                                                  //
//                                                                                      //
//  Eingabe: · timer: Zeiger auf die Zeitgeber-Struktur.                                //
//  Ausgabe: 'true', wenn der Zeitgeber abgelaufen ist.                                 //
//////////////////////////////////////////////////////////////////////////////////////////
bool
TIMER_HasExpired(Timer* timer)
{
    return TIMER_Remaining(timer) == 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  TIMER_RestartIfExpired(): Die Funktion ist eine "convenient function" für die       //
//  ------------------------  Hintereinanderausführung von TIMER_HasExpired() und       //
//  TIMER_Restart(). Sie fasst damit die folgende Sequenz zusammen:                     //
//                                                                                      //
//       if (TIMER_HasExpired(...))           if (TIMER_RestartIfExpired(...))          //
//       {   TIMER_Restart(...);              {                                         //
//           ...                        zu         ...                                  //
//       }                                    }                                         //
//                                                                                      //
//  Eingabe: · timer: Zeiger auf die Zeitgeber-Struktur.                                //
//  Ausgabe: 'true', wenn der Zeitgeber abgelaufen war und neu gestartet wurde.         //
//////////////////////////////////////////////////////////////////////////////////////////
bool
TIMER_RestartIfExpired(Timer* timer)
{
    if (!TIMER_HasExpired(timer)) return false;
    TIMER_Restart(timer);
    return true;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  TIMER_Elapsed(): Liefert die Zeit, die der Zeitgeber bereits läuft. Ist der Zeit-   //
//  ---------------  geber abgelaufen, liefert die Funktion den gleichen Wert wie       //
//  TIMER_Duration().                                                                   //
//                                                                                      //
//  Eingabe: · timer: Zeiger auf die Zeitgeber-Struktur.                                //
//  Ausgabe: Die verstrichene Zeit: 0 .. TIMER_Duration() [s|ms].                       //
//////////////////////////////////////////////////////////////////////////////////////////
UInt16
TIMER_Elapsed(Timer* timer)
{
    return timer->duration - TIMER_Remaining(timer);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  TIMER_Remaining(): Liefert die noch verbleibende Zeit bis zum Ablauf des Zeit-      //
//  -----------------  gebers. Ist der Zeitgeber abgelaufen, liefert die Funktion 0.    //
//                                                                                      //
//  Eingabe: · timer: Zeiger auf die Zeitgeber-Struktur.                                //
//  Ausgabe: Verbleibende Zeit: 0 .. TIMER_Duration() [s|ms].                           //
//////////////////////////////////////////////////////////////////////////////////////////
UInt16
TIMER_Remaining(Timer* timer)
{
    if (!timer->hasExpired)
    {   UInt16 passedTicks = Ticks(timer) - timer->ticksStamp;
        if (passedTicks < timer->duration) return timer->duration - passedTicks;
        timer->hasExpired = true;
    }
    return 0;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  TIMER_Duration(): Liefert die mittels TIMER_Start() oder per TIMER_..()-Struktur    //
//  ----------------  initialisierte Dauer des Zeitgebers.                              //
//                                                                                      //
//  Eingabe: · timer: Zeiger auf die Zeitgeber-Struktur.                                //
//  Ausgabe: Dauer des Zeitgebers: 0 .. 15000 [s|ms].                                   //
//////////////////////////////////////////////////////////////////////////////////////////
UInt16
TIMER_Duration(const Timer* timer)
{
    return timer->duration;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  TIMER_SetBase(): Neben der Initialisierungsstruktur TIMER_..() ist dies die zweite  //
//  ---------------  Möglichkeit, die Basis des Zeitgebers festzulegen. Mögliche Werte  //
//  für <base> sind TIMERBASE_MILLISECONDS und TIMERBASE_SECONDS. Zu Zeitbasis siehe    //
//  auch TIMER_Start(). Nach dem Setzen der Zeitbasis ist der Zeitgeber abgelaufen.     //
//                                                                                      //
//  Eingabe: · timer: Zeiger auf die Zeitgeber-Struktur.                                //
//           · base:  Zeitbasis als Aufzählungstyp 'TimerBase'.                         //
//  Ausgabe: -                                                                          //
//////////////////////////////////////////////////////////////////////////////////////////
void
TIMER_SetBase(Timer* timer, TimerBase base)
{
    timer->hasExpired = true;
    timer->base = base;
}

//////////////////////////////////////////////////////////////////////////////////////////
//  TIMER_Base(): Liefert die mittels TIMER_SetBase() oder per Initialisierungsstruktur //
//  ------------  TIMER_..()-Struktur festgelegte Basis des Zeitgebers.                 //
//                                                                                      //
//  Eingabe: · timer: Zeiger auf die Zeitgeber-Struktur.                                //
//  Ausgabe: Zeitbasis als Aufzählungstyp 'TimerBase'.                                  //
//////////////////////////////////////////////////////////////////////////////////////////
TimerBase
TIMER_Base(const Timer* timer)
{
    return timer->base;
}

//////////////////////////////////////////////////////////////////////////////////////////
